# -*- coding: utf-8 -*-
"""Simple fit to a gaussian distribution

In this example a fit is performed to a simple gaussion distribution.

Observables can be initialised by a list with the column name / variable name as first argument, followed
by the range and/or with the initial value and range:
x = ('x', -3, 3)
x = ('mass', -3, 0.02, 3)

Parameters are initialised with a tuple: sigma=(0,1) or again including a starting parameter: sigma=(0.01, 0, 1)
The order here is not important.

"""

import pyminufit as mnf
import numpy as np


data = np.append(
    np.random.random_sample(1000) * 10 + 745, np.random.normal(750, 1, 1000)
)

x = mnf.create_real_var(("mass", 745, 755), unit="GeV")

pdf_sig = mnf.Gauss(x, mean=(745, 755), sigma=(0.1, 1, 2), title="Signal")
pdf_bkg = mnf.Chebyshev(x, order=2, title="Background")

pdf = pdf_sig + pdf_bkg

pdf.fit(data)
pdf.plot(data, "02_add_pdf.png", dpi=150)
pdf.get()



# -*- coding: utf-8 -*-
"""Simple fit to a gaussian distribution

In this example a fit is performed to a simple gaussion distribution.

Observables can be initialised by a list with the column name / variable name as first argument, followed
by the range and/or with the initial value and range:
x = ('x', -3, 3)
x = ('mass', -3, 0.02, 3)

Parameters are initialised with a tuple: sigma=(0,1) or again including a starting parameter: sigma=(0.01, 0, 1)
The order here is not important.

All parameters and observables can also be initialised by a ROOT.RooRealVar.

"""

from pyminufit.models import Gauss
import numpy as np

data = np.random.normal(0, 1, 1000)

pdf = Gauss(("x", -3, 3), mean=(-1, 0, 1))
pdf.fit(data)
pdf.plot(data, "01_simple_fit.png")

pdf.get()



from __future__ import annotations

import pytest

from pyminufit.utils import AttrDict, is_iterable


def test_is_iterable():
    assert is_iterable([])
    assert is_iterable((1, 2, 3))
    assert is_iterable("hello")
    assert not is_iterable(123)
    assert not is_iterable(None)


def test_AttrDict():
    data = {"a": 1, "b": 2, "c": 3}
    attr_dict = AttrDict(data)

    # Test item lookup
    assert attr_dict["a"] == 1
    assert attr_dict["b"] == 2
    assert attr_dict["c"] == 3

    # Test attribute lookup
    assert attr_dict.a == 1
    assert attr_dict.b == 2
    assert attr_dict.c == 3

    # Test modifying item
    attr_dict["a"] = 10
    assert attr_dict.a == 10

    # Test modifying attribute
    attr_dict.b = 20
    assert attr_dict["b"] == 20

    # Test adding new item
    attr_dict.d = 30
    assert attr_dict["d"] == 30
    assert attr_dict.d == 30

    # Test deleting item
    del attr_dict["c"]
    with pytest.raises(KeyError):
        attr_dict["c"]

    # Test deleting attribute
    del attr_dict.b
    with pytest.raises(AttributeError):
        attr_dict.b



"""
Copyright (c) 2024 Simon. All rights reserved.

pyminufit: A great package.
"""

from __future__ import annotations

from .models import Chebyshev, Gauss, Normal
from .observables import RealVar, create_real_var
from .version import version as __version__

__all__ = ["Gauss", "Normal", "Chebyshev", "create_real_var", "RealVar", "__version__"]



"""
Observables

"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional, Union


@dataclass
class RealVar:
    """
    Represents a real-valued variable with associated properties.

    Attributes:
        name (str): The name of the variable.
        value (float): The value of the variable.
        error (Optional[float]): The error associated with the variable (if available).
        lwb (Optional[float]): The lower bound of the variable (if available).
        upb (Optional[float]): The upper bound of the variable (if available).
        unit (Optional[str]): The unit of the variable (if available).
        is_constant (bool): Indicates whether the variable is constant or not.

    Methods:
        __float__(): Returns the value of the variable as a float.
        __call__(): Returns the value of the variable as a float.
        _repr_html_(): Returns an HTML representation of the variable.

    """

    name: str
    value: float
    error: Optional[float]
    lwb: Optional[float]
    upb: Optional[float]
    unit: Optional[str]
    is_constant: bool = False

    def __float__(self) -> float:
        return float(self.value)

    def __call__(self) -> float:
        return float(self.value)

    def _repr_html_(self) -> str:
        val = f"<b>{self.name}</b>: {self.value:.3g}"
        val += f" ± {self.error:.3g}" if self.error else ""
        return val + f" {self.unit}"


def extract_from_list(var: Union[list, tuple]) -> tuple:  # type: ignore[type-arg]
    """Extract name, min, mean and max from a list

    Args:
        var (list or tuple): List in a format like ['x', -1, 1] or (-2, 0, 3)

    Returns:
        Tuple containing name, value, min, max

    """
    var = list(var)
    val = lwb = upb = None
    name = "x"

    assert len(var) >= 2, "Please use a format like ['x', -1, 1] or (-2, 0, 3)"

    if isinstance(var[0], str):
        name = var[0]
        var = var[1:]
    var = sorted(var)
    if len(var) == 2:
        lwb, upb = var
        val = lwb + upb
        val /= 2.0
    if len(var) == 3:
        lwb, val, upb = var

    return name, val, lwb, upb


def create_real_var(
    var: Optional[Any] = None,
    name: Optional[str] = None,
    lwb: Optional[float] = None,
    upb: Optional[float] = None,
    value: Optional[float] = None,
    unit: str = "",
    error: Optional[float] = None,
) -> RealVar:
    """Create a RealVar object

    Args:
        var (list or tuple): List in a format like ['x', -1, 1] or (-2, 0, 3)
        name (str): Name of the variable
        lwb (float): Lower bound
        upb (float): Upper bound
        value (float): Value
        unit (str): Unit

    Returns:
        RealVar object

    """
    if isinstance(var, RealVar):
        return var

    name_override = name
    if var:
        override_bounds = False
        if isinstance(var, float):
            override_bounds = True
            var = (var, var, var)
        name, value, lwb, upb = extract_from_list(var)
        if override_bounds:
            lwb = upb = None
    if name_override:
        name = name_override
    return RealVar(name, value, error, lwb, upb, unit)  # type: ignore[arg-type]



from __future__ import annotations
import pytest
from pyminufit.observables import create_real_var


def test_create_real_var():
    # Test case 1: Testing with var=None
    result = create_real_var()
    assert result.name == None
    assert result.value is None
    assert result.lwb is None
    assert result.upb is None
    assert result.unit == ""

    # Test case 2: Testing with var=['x', -1, 1]
    result = create_real_var(var=["x", -1, 1])
    assert result.name == "x"
    assert result.value == 0
    assert result.lwb == -1
    assert result.upb == 1
    assert result.unit == ""

    # Test case 3: Testing with var=(-2, 0, 3)
    result = create_real_var(var=(-2, 0, 3))
    assert result.name == "x"
    assert result.value == 0
    assert result.lwb == -2
    assert result.upb == 3
    assert result.unit == ""

    # Test case 4: Testing with custom values
    result = create_real_var(name="y", lwb=-10, upb=10, value=5, unit="m")
    assert result.name == "y"
    assert result.value == 5
    assert result.lwb == -10
    assert result.upb == 10
    assert result.unit == "m"
    assert float(result) == 5



"""Automatic combination of PDFs

This module handles the automatic combination of PDFs.
A combination can be a product or addition of two PDFs.

Also convolution is supported.

"""

from __future__ import annotations

from typing import Any, Optional, Tuple, Union

from iminuit import Minuit
from iminuit.cost import ExtendedUnbinnedNLL
from numpy.typing import ArrayLike

from pyminufit.observables import create_real_var

from .pdf import Pdf
from .utils import AttrDict


class AddPdf(Pdf):
    """Add PDF class, for generic addition of pdfs

        This class combines two PDF classes into an new PDF normed by normalisation factors.
        It is the generic way to fit multiple components in a fit.

        Examples:
            Two models of clas ``PDF`` can easily combined automatically by calling:

            .. code-block:: python

                add_pdf = pdf1 + pdf2

            If more than two models are to be combined:

            .. code-block:: python

                 add_pdf = AddPdf([pdf1, pdf2, pdf3])

        Attributes:
            pdfs (AttrDict): dict of pdfs in the object
            norms (AttrDict): dict of normalisations
            primary_pdf (Pdf): primary pdf
    def __init__(self, pdfs: Optional[List[Pdf]] = None, name: Optional[str] = None, **kwds) -> None:

        Todo:
            * Allow for absolute and relative normalisations.

    """

    def __init__(
        self, pdfs: Optional[list[Pdf]] = None, name: Optional[str] = None, **kwds: Any
    ) -> None:
        self.pdfs = AttrDict()
        self.primary_pdf: Optional[Pdf] = None
        self.norms = AttrDict()
        self._external_norms = AttrDict()

        if pdfs is not None:
            if name is None:
                name = "_plus_".join(pdf.name for pdf in pdfs)
        else:
            pdfs = []
            if name is None:
                name = "AddPdf"
        super().__init__(name=name, **kwds)

        for pdf in pdfs:
            self.add(pdf)

        self.use_extended = True

    def add(self, pdf: Pdf) -> None:
        """Add a PDF to the AddPdf object

        Args:
            pdf (Pdf): pdf to be added to the object

        """
        if not self.primary_pdf:
            self.primary_pdf = pdf

        # Check for duplicate names
        if pdf.name in self.pdfs:
            msg = f"PDF with name {pdf.name} already exists"
            raise ValueError(msg)

        self.pdfs[pdf.name] = pdf
        self.norms["n_" + pdf.name] = 1.0
        self.init_pdf()

    def constrain_norm(self, pdf: Union[Pdf, str], normalization: float) -> None:
        """Constrain a norm with an external normalisation

        Args:
            pdf (:obj:`PDF` or :obj:`str`): PDF object from the AddPdf or name of corresponding PDF
            normalization (:obj:`ROOT.RooAbsReal`): New normalisation to be used in the ROOT.RooAddPdf initialisation

        """
        assert isinstance(pdf, (Pdf, str)), "please specify pdf"
        if isinstance(pdf, Pdf):
            pdf = pdf.name
        assert pdf in self.pdfs, "Pdf not found"

        # assert isinstance(normalization, ROOT.RooAbsReal), "Please provide ROOT.RooAbsReal type as normalisation"

        self._external_norms[pdf] = normalization
        self.init_pdf()  # reset the pdf and change corresponding normalisation

    def init_pdf(self) -> None:
        """Initialises the PDF object

        This method is called when a new PDF is added to the object.

        """
        for pdf_name, pdf in self.pdfs.items():
            if pdf_name not in self._external_norms:
                norm_var = create_real_var(("n_" + pdf_name, 10, 0, 100000000))
            else:
                norm_var = self._external_norms[pdf_name]
            self.norms[pdf_name] = norm_var
            self.parameters["n_" + pdf_name] = norm_var
            self.observables.update(pdf.observables)

            for param_name, param in pdf.parameters.items():
                self.parameters[pdf_name + "_" + param_name] = param

    def _fit(self, data: ArrayLike, **kwds: Any) -> Minuit:
        self.cost = ExtendedUnbinnedNLL(data, self.density)
        m = Minuit(
            self.cost,
            **{k: v.value for k, v in self.parameters.items()},
            name=tuple(self.parameters.keys()),
        )
        self._set_limits(m)
        m.migrad(**kwds)
        m.hesse()
        return m

    def density(self, x, *args) -> Tuple[float, float]:  # type: ignore[no-untyped-def]
        kwargs = dict(zip(self.parameters.keys(), args))
        norm_names = ["n_" + pdf_name for pdf_name in self.pdfs]
        norm = sum([kwargs[n] for n in norm_names])
        component_sum = 0
        for pdf_name, pdf in self.pdfs.items():
            pdf_params_dict = {k: kwargs[pdf_name + "_" + k] for k in pdf.parameters}
            component_sum += kwargs["n_" + pdf_name] * pdf.pdf(x, **pdf_params_dict)
        return norm, component_sum

    def pdf(self, x: ArrayLike, *args, **kwds) -> ArrayLike:  # type: ignore[no-untyped-def]
        kwargs = {k: v.value for k, v in self.parameters.items()}
        norm, component_sum = self.density(x, *kwargs.values())
        return component_sum / norm



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from numpy.typing import ArrayLike, NDArray

HATCHES = [
    "",
    "///",
    r"\\\ ",
    "xxx",
    "--",
    "++",
    "o",
    ".+",
    "xx",
    "//",
    "*",
    "O",
    ".",
]
DEFAULT_COLORS = [
    "#E24A33",
    "#348ABD",
    "#988ED5",
    "#777777",
    "#FBC15E",
    "#8EBA42",
    "#FFB5B8",
]


@dataclass
class Hist:
    """
    Represents a histogram.

    Attributes:
        n (int): The number of bins in the histogram.
        xe (numpy.ndarray): The bin edges of the histogram.

    Methods:
        __post_init__(self) -> None: Initializes the histogram object and calculates additional attributes.
    """

    n: int
    xe: NDArray

    def __post_init__(self) -> None:
        self.ne = self.n**0.5
        self.cx = 0.5 * (self.xe[1:] + self.xe[:-1])
        self.dx = np.diff(self.xe)
        self.xlim = (self.xe[0], self.xe[-1])


class Plotter:
    def __init__(
        self,
        model: Any,
        data: ArrayLike,
        nbins: Optional[int] = None,
        xrange: Optional[Tuple[float, float]] = None,
        ax: Optional[List[plt.Axes]] = None,
        cfg: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Initialize the Plotter object.

        Args:
            model: The model used for fitting.
            data: The data to be plotted.
            nbins: The number of bins for the histogram. If not provided, it is calculated based on the data length.
            xrange: The range of the x-axis. If not provided, it is determined based on the minimum and maximum values of the data.
            ax: The list of matplotlib Axes objects to plot on. If not provided, a new Axes object will be created.
            cfg: Additional configuration options for the plot. If not provided, an empty dictionary is used.

        Returns:
            None
        """
        self.cfg = cfg or {}
        self.model = model
        self.data = data
        self.xrange = xrange or (np.min(data), np.max(data))
        self.nbins = nbins or int(2 * len(data) ** 0.333)
        self.h = Hist(*np.histogram(data, bins=self.nbins, range=self.xrange))
        self.scale = len(data) * self.h.dx[0]
        self.ax: List[plt.Axes] = ax or []
        self.parts: Dict[str, Callable[[], None]] = {
            "hist": self.plot_hist,
            "pdf": self.plot_pdf,
            "pull": self.plot_pull,
            "residuals": self.plot_residuals,
            "legend": self.plot_legend,
            "components": self.plot_components,
            "axes_labels": self.plot_axes_labels,
            "distribution": self.plot_distribution,
        }

    def setup_axis(
        self, ax: Optional[Union[plt.Axes, List[plt.Axes]]], components: List[str]
    ) -> List[plt.Axes]:
        if not ax:
            nx = 1
            gridspec_kw = None
            if components and ("pull" in components):
                nx = 2
                gridspec_kw = {"height_ratios": [4, 1]}
            _, ax = plt.subplots(
                nx,
                1,
                sharex=True,
                figsize=self.cfg.get("figsize", (4, 4)),
                gridspec_kw=gridspec_kw,
            )
            if nx > 1:
                ax = list(ax)
        if not isinstance(ax, list):
            ax = [ax]
        return ax  # type: ignore[no-any-return]

    def plot(
        self,
        components: Optional[List[str]] = None,
        filename: Optional[str] = None,
        *args: Any,
        **kwargs: Any,
    ) -> Plotter:
        if not components:
            # check if self.model has attribute pdfs
            if hasattr(self.model, "pdfs"):
                components = [
                    "hist",
                    "pdf",
                    "pull",
                    "components",
                    "legend",
                    "axes_labels",
                ]
            else:
                components = ["hist", "pdf", "axes_labels", "distribution"]
        self.ax = self.setup_axis(None, components)
        for c in components:
            self.parts[c]()
        if len(self.ax) > 1:
            plt.subplots_adjust(hspace=0)
        if filename:
            plt.savefig(filename, *args, **kwargs, bbox_inches="tight")
        return self

    def plot_hist(self) -> None:
        self.ax[0].errorbar(self.h.cx, self.h.n, self.h.ne, self.h.dx / 2.0, fmt=".k")
        self.ax[0].set_xlim(min(self.data), max(self.data))
        self.ax[0].set_ylim(0, None)

    def plot_pdf(self) -> None:
        xx = np.linspace(*self.h.xlim, 1000)
        self.ax[0].plot(xx, self.scale * self.model(xx), color="black")

    def plot_pull(self) -> None:
        yerr = self.h.ne
        yerr[yerr == 0] = np.inf
        pull = (self.h.n - self.scale * self.model(self.h.cx)) / yerr

        self.ax[1].bar(self.h.cx, pull, width=self.h.dx, alpha=0.5, color="gray")
        self.ax[1].hlines(0, *self.h.xlim, linestyles="dashed", color="black")
        self.ax[1].tick_params(which="both", top=True, right=True)
        self.ax[1].set_ylim(-2, 2)
        self.ax[1].set_yticks([-1, 1])

    def plot_residuals(self) -> None:
        self.ax[1].errorbar(
            self.h.cx,
            self.h.n - self.scale * self.model(self.h.cx),
            self.h.ne,
            fmt="ok",
        )
        self.ax[1].hlines(0, *self.h.xlim, linestyles="dashed", color="black")

    def plot_components(self) -> None:
        from pyminufit.composites import AddPdf

        if not isinstance(self.model, AddPdf):
            return
        ax = self.ax[0]

        for i, pdf_name in enumerate(self.model.pdfs):
            pdf = self.model.pdfs[pdf_name]

            pdf_norm = self.model.norms[pdf.name].value / sum(
                self.model.norms[pdf].value for pdf in self.model.pdfs
            )
            xx = np.linspace(*self.h.xlim, 1000)
            yy = self.scale * pdf_norm * pdf(xx)
            current_color = (
                ax._get_lines.get_next_color()
                if self.cfg.get("override_color", False)
                else DEFAULT_COLORS[i % len(DEFAULT_COLORS)]
            )
            ax.plot(xx, yy, color=current_color)
            ax.fill_between(
                xx,
                yy,
                alpha=0.5,
                hatch=HATCHES[i % len(HATCHES)],
                facecolor="none" if i > 0 else current_color,
                edgecolor=current_color,
                label=pdf.title,
            )

    def plot_legend(self) -> None:
        self.ax[0].legend(**self.cfg.get("legend", {}), frameon=False)

    def plot_axes_labels(self) -> None:
        binsize = self.h.dx[0]
        obs_name = self.model.observable.name
        obs_unit = self.model.observable.unit or ""
        per_bin = f"{binsize:.1g}"
        if obs_unit != "":
            per_bin = f" ( {per_bin} {obs_unit} )"
        self.ax[0].set_ylabel(f"Entries / {per_bin}")
        if len(self.ax) > 1:
            self.ax[1].set_ylabel("Pull")

        if len(obs_unit) > 0:
            obs_unit = f" [{obs_unit}]"
        self.ax[-1].set_xlabel(f"{obs_name}{obs_unit}")

    def plot_distribution(self) -> None:
        self.ax[0].plot(self.data, np.zeros_like(self.data), "|", alpha=0.1)



"""PDF base class"""

from __future__ import annotations

from typing import Any, Dict, Optional

from iminuit import Minuit
from numpy.typing import ArrayLike

from .observables import RealVar, create_real_var
from .plotting import Plotter
from .utils import AttrDict, ClassLoggingMixin


class Pdf(ClassLoggingMixin):
    """PDF base class"""

    def __init__(
        self,
        name: str,
        observables: Optional[Any] = None,
        title: Optional[str] = None,
        **kwds: Any,
    ) -> None:
        """
        Initialize a PDF object.

        Args:
            name (str): The name of the PDF.
            observables (Optional[Any]): A list of observables.
            title (Optional[str]): The title of the PDF. If not provided, it defaults to the name.
            **kwds (Any): Additional keyword arguments.

        """
        super().__init__(**kwds)
        self.name = name
        self.title = title or name
        self.observables = AttrDict()
        if observables:
            for observable in observables:
                self.add_observable(observable)
        self.parameters = AttrDict()
        self.parameters_names = AttrDict()
        self.cost = None
        self._init_pdf()

    def _init_pdf(self) -> None:
        """Initiate attributes for parameters"""
        for p in self.parameters:
            setattr(self, p, self.parameters[p])

    def add_parameter(
        self,
        param_var: Any,
        param_name: Optional[str] = None,
        final_name: Optional[str] = None,
        **kwds: Any,
    ) -> RealVar:
        """
        Adds a parameter to the PDF object.

        Args:
            param_var (Any): The value of the parameter.
            param_name (Optional[str]): The name of the parameter. If not provided, a final name must be provided.
            final_name (Optional[str]): The final name of the parameter. If provided, it will be used as the name instead of param_name.
            **kwds (Any): Additional keyword arguments to be passed to create_real_var.

        Returns:
            RealVar: The added parameter.

        """
        if final_name is None:
            assert (
                param_name is not None
            ), "Please provide a final name for the parameter"
            name = self.name + "_" + param_name
        else:
            name = final_name
        assert isinstance(param_name, str), "Please provide a name for the parameter"
        param = create_real_var(param_var, name=name, **kwds)
        self.parameters[param_name] = param
        self.parameters_names[param_name] = name
        setattr(self, param_name, param)
        return self.parameters[param_name]  # type: ignore[no-any-return]

    @property
    def observable(self) -> RealVar:
        """Return the first observable"""
        if len(self.observables) == 0:
            msg = "No observables defined"
            raise ValueError(msg)
        if len(self.observables) > 1:
            self.warn("Multiple observables defined, returning the first one")
        return next(iter(self.observables.values()))  # type: ignore[no-any-return]

    def add_observable(self, observable_var: Any, **kwds: Any) -> RealVar:
        """Add an observable to the PDF"""
        if isinstance(observable_var, (list, tuple)) and not isinstance(
            observable_var[0], str
        ):
            self.warn("WARNING : choosing automatic variable name 'x'")

        observable = create_real_var(observable_var, **kwds)
        name = observable.name
        self.observables[name] = observable
        return self.observables[name]  # type: ignore[no-any-return]

    def fit(self, data: ArrayLike, **kwds: Any) -> Minuit:
        """Fit the PDF to the data"""
        self.logger.debug("Fitting")
        fit_res = self._fit(data, **kwds)
        self._update_parameters(fit_res)
        return fit_res

    def _fit(self, data: ArrayLike, **kwds: Any) -> Minuit:
        """Fit logic implemented in the models"""
        # pylint: disable=unused-argument
        msg = "Please implement the fit method"
        raise NotImplementedError(msg)

    def _update_parameters(self, m: Minuit) -> None:
        for p in self.parameters:
            self.parameters[p].value = m.values[p]
            self.parameters[p].error = m.errors[p]
            setattr(self, p, self.parameters[p])

    def _set_limits(self, m: Minuit) -> None:
        for p in self.parameters:
            m.limits[p] = (self.parameters[p].lwb, self.parameters[p].upb)

    def pdf(self, x: ArrayLike, *args, **kwds) -> ArrayLike:  # type: ignore[no-untyped-def]
        """PDF logic implemented in the models"""
        # pylint: disable=unused-argument
        return x

    def evaluate(self, x: ArrayLike) -> ArrayLike:
        """
        Evaluates the probability density function (PDF) at the given values of x.
        Args:
            x (ArrayLike): The values at which to evaluate the PDF.
        Returns:
            ArrayLike: The values of the PDF at the given values of x.
        """
        parameter_values = {p: self.parameters[p].value for p in self.parameters}
        return self.pdf(x, **parameter_values)

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return self.evaluate(*args, **kwds)

    def __add__(self, other: Pdf) -> Pdf:
        from .composites import AddPdf

        return AddPdf([self, other])

    def plot(
        self,
        data: ArrayLike,
        filename: Optional[str] = None,
        cfg: Optional[Dict[str, Any]] = None,
        *args: Any,
        **kwargs: Any,
    ) -> Plotter:
        """Plot the PDF"""
        return Plotter(self, data, cfg=cfg).plot(None, filename, *args, **kwargs)

    def get(self, parameter: Optional[str] = None, as_ufloat: bool = False) -> Any:
        """Get one of the fitted parameter or print all if None is set

        Args:
            parameter (str): name of the parameter
            as_ufloat (bool, optional): If true return ufloat object, else tuple

        Returns:
            :obj:`tuple` or :obj:`ufloat` mean and error of parameter

        """

        if parameter is None:
            for m in self.parameters:
                self.info(f"{m:18} ==> {self._get_parameter(m, True)}")
        else:
            return self._get_parameter(parameter, as_ufloat)
        return self.parameters

    def _get_parameter(self, param: str, as_ufloat: bool = False) -> Any:
        """Internal getter for parameter values

        Args:
            param: Parameter name
            as_ufloat: Return ufolat object

        Returns:
            :obj:`tuple` or :obj:`ufloat` mean and error of parameter
        """
        mes = self.parameters[param]
        val = mes.value

        # Now catch RooFormulaVar
        try:
            err = mes.error
        except AttributeError:
            err = 0
        if not as_ufloat:
            return val, err
        try:
            from uncertainties import ufloat

            return ufloat(val, err)
        except ImportError:
            return val, err

    def __getitem__(self, parameter_name: str) -> Any:
        return self._get_parameter(parameter_name)



"""
Some utility constructs.

"""

from __future__ import annotations

import collections
import logging
from typing import Any


class ClassLoggingMixin:
    """Mixin class that enables logging for instances of a specific class"""

    def __init__(self, *args: Any, **kwds: Any) -> None:
        """Initialise the logger instance"""
        super().__init__(*args, **kwds)
        self.logger = logging.getLogger(self.__class__.__name__)

    def debug(self, msg: str) -> None:
        self.logger.debug(msg)

    def info(self, msg: str) -> None:
        self.logger.info(self, msg)

    def warn(self, msg: str) -> None:
        self.logger.warning(msg)

    def error(self, msg: str) -> None:
        self.logger.error(msg)

    def unknown_error(self) -> None:
        self.logger.error("Unknown Error occurred o_O")

    @staticmethod
    def setup_basic_config() -> None:
        """Setup basic logging configuration"""
        logging.basicConfig(
            level=logging.DEBUG,
            format="%(name)-18s \t %(levelname)-8s %(message)s",
            datefmt="%m-%d %H:%M",
        )


class AttrDict(dict):  # type: ignore[type-arg]
    """Dictionary which items can also be addressed by attribute lookup in addition to item lookup"""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.__dict__ = self


def is_iterable(obj: Any) -> bool:
    """Check if an object is iterable

    Args:
        obj (object): Generic object

    Returns:
        True if is able to use python iterations

    """
    return isinstance(obj, collections.abc.Iterable)



"""Fit models predefined"""

from __future__ import annotations

from typing import Any, Dict, Sequence, Union

import numpy as np
from iminuit import Minuit
from iminuit.cost import UnbinnedNLL
from numpy.polynomial.chebyshev import chebval
from numpy.typing import ArrayLike
from scipy.stats import norm

from .observables import RealVar
from .pdf import Pdf

__all__ = ["Gauss", "Normal", "Chebyshev"]


class Gauss(Pdf):
    """Standard gaussian"""

    def __init__(
        self,
        observable: Union[Sequence[Any], RealVar],
        mean: Union[Sequence[Any], RealVar] = (-1, 0, 1),
        sigma: Union[Sequence[Any], RealVar] = (0, 1),
        name: str = "gauss",
        *args: Any,
        **kwds: Any,
    ):
        super().__init__(name, *args, **kwds)

        self.add_observable(observable)
        self.add_parameter(mean, "mean")
        self.add_parameter(sigma, "sigma")
        self._init_pdf()

    def _fit(self, data: ArrayLike, **kwds: Any) -> Minuit:
        self.cost = UnbinnedNLL(data, self.pdf)
        m = Minuit(
            self.cost,
            mean=self.parameters["mean"].value,
            sigma=self.parameters["sigma"].value,
        )
        self._set_limits(m)
        m.migrad(**kwds)
        m.hesse()
        self._update_parameters(m)
        return m

    def pdf(self, x, mean, sigma):  # type: ignore[no-untyped-def]
        return norm.pdf(x, mean, sigma)


# Add alias
Normal = Gauss


class Chebyshev(Pdf):
    """Chebyshev"""

    def __init__(
        self,
        observable: Union[Sequence[Any], RealVar],
        order: int = 2,
        name: str = "chebyshev",
        do_normalize: bool = True,
        *args: Any,
        **kwds: Any,
    ):
        super().__init__(name, *args, **kwds)
        self.do_normalize = do_normalize
        self.add_observable(observable)
        for i in range(order):
            self.add_parameter((-1, 1), f"c{i}")
        self._init_pdf()

    def pdf(self, x, *args, **kwargs: Dict[str, Any]):  # type: ignore[no-untyped-def]
        coeffs = args if args else tuple(kwargs.values())

        c = np.array(list(coeffs))
        norm = 1.0

        x_range = np.linspace(self.observable.lwb, self.observable.upb, 1000)
        y_range = chebval(x_range, c)
        shift = -1 * min(np.min(y_range), 0)
        y_range += shift

        if self.do_normalize:
            # integrated_c = chebint(c, m=1)  # type: ignore[no-untyped-call]
            # norm = chebval(self.observable.upb, integrated_c) - chebval(self.observable.lwb, integrated_c)  # type: ignore[no-untyped-call]
            norm = np.trapz(y_range, x_range)  # integrate over the range
            if np.isnan(norm) or norm == 0:
                norm = 1.0

        y = chebval(x, c) + shift
        return y / norm

    def _fit(self, data: ArrayLike, **kwds: Any) -> Minuit:
        self.cost = UnbinnedNLL(data, self.pdf)
        m = Minuit(
            self.cost,
            **{k: v.value for k, v in self.parameters.items()},
            name=tuple(self.parameters.keys()),
        )
        self._set_limits(m)
        m.migrad(**kwds)
        m.hesse()
        self._update_parameters(m)
        return m



import pytest
from pyminufit.models import Gauss


def test_Gauss_fit():
    # Test case 1: Testing with default parameters
    gauss = Gauss(observable=(-1, 1))
    data = [0.5, 1.0, 1.5, 2.0, 2.5]
    result = gauss.fit(data)
    assert result.valid
    assert gauss.parameters["mean"].value == pytest.approx(1.5, abs=1e-0)
    assert gauss.parameters["sigma"].value == pytest.approx(0.5, abs=1e-0)

    # Test case 2: Testing with custom parameters
    gauss = Gauss(observable=("x", -1, 1), mean=(-2, -1, 0), sigma=(1, 2))
    data = [-1.0, 0.0, 1.0]
    result = gauss.fit(data)
    assert result.valid
    assert gauss.parameters["mean"].value == pytest.approx(0.0, abs=1e-0)
    assert gauss.parameters["sigma"].value == pytest.approx(1.0, abs=1e-0)
